///	\file gsc16ai64.c
///	\brief File contains the initialization routine and various register
/// read/write
///<		operations for the General Standards 16bit, 32 channel ADC
///< modules. \n
///< For board info, see
///<    <a
///<    href="http://www.generalstandards.com/view-products2.php?BD_family=16ai64ssc">GSC
///<    16AI64SSC Manual</a>

#include "gsc_adc_common.h"
#include "gsc18ai32.h"

// *****************************************************************************
/// \brief Routine to initialize GSC 16bit, 32 channel ADC modules
///     @param[in,out] *pHardware Pointer to global data structure for storing
///     I/O
///<            register mapping information.
///     @param[in] *adcdev PCI address information passed by the mapping code in
///     map.c
///	@return Status from board enable command.
// *****************************************************************************
int
gsc18ai32Init( CDS_HARDWARE* pHardware, struct pci_dev* adcdev )
{
    static unsigned int pci_io_addr; /// @param pci_io_addr Bus address of PCI
                                     /// card I/O register.
    int devNum; /// @param devNum Index into CDS_HARDWARE struct for adding
                /// board info.
    char* _adc_add; /// @param *_adc_add ADC register address space
    int   pedStatus; /// @param pedStatus Status return from call to enable
                   /// device.
    volatile GSA_ADC_REG* adc18Ptr;
    int                   autocal = 0;
    int                   chans = 2;

    /// Get index into CDS_HARDWARE struct based on total number of ADC cards
    /// found by mapping routine in map.c
    devNum = pHardware->adcCount;
    /// Enable the module.
    pedStatus = pci_enable_device( adcdev );
    /// Enable device to be DMA master.
    pci_set_master( adcdev );
    /// Get the PLX chip address
    pci_read_config_dword( adcdev, PCI_BASE_ADDRESS_0, &pci_io_addr );
    printk( "pci0 = 0x%x\n", pci_io_addr );
    /// Map module DMA space directly to computer memory space.
    _adc_add = ioremap_nocache( (unsigned long)pci_io_addr, 0x200 );
    /// Map the module DMA control registers via PLX chip registers
    adcDma[ devNum ] = (PLX_9056_DMA*)_adc_add;

    /// Get the ADC register address
    pci_read_config_dword( adcdev, PCI_BASE_ADDRESS_2, &pci_io_addr );
    printk( "pci2 = 0x%x\n", pci_io_addr );
    /// Map the module control register so local memory space.
    _adc_add = ioremap_nocache( (unsigned long)pci_io_addr, 0x200 );
    printk( "ADC 18 I/O address=0x%x  0x%lx\n", pci_io_addr, (long)_adc_add );
    /// Set global ptr to control register memory space.
    adc18Ptr = (GSA_ADC_REG*)_adc_add;
    _adcPtr[ devNum ] = (GSA_ADC_REG*)_adc_add;
    printk( "ADC  pointer init  at 0x%lx\n", (long)_adcPtr[ devNum ] );

    printk( "BCR = 0x%x\n", adc18Ptr->BCR );
    /// Reset the ADC board
    adc18Ptr->BCR |= GSAF_RESET;
    do
    {
    } while ( ( adc18Ptr->BCR & GSAF_RESET ) != 0 );

    /// Write in a sync word
    adc18Ptr->SMUW = 0x0000;
    adc18Ptr->SMLW = 0x0000;

    /// Set ADC to 64 channel = 32 differential channels
    // Don't need this for this board

#ifdef ADC_EXTERNAL_SYNC
    // This is for the external connector use
    adc18Ptr->BCR |= ( GSAF_ENABLE_X_SYNC ); // 0x80
    adc18Ptr->AUX_SIO |= 0x80;
#endif

    /// Set sample rate close to 16384Hz
    /// Unit runs with external clock, so this probably not necessary
    adc18Ptr->RAG =
        (unsigned int)( GSC18AI32_OSC_FREQ / ( UNDERSAMPLE * IOP_IO_RATE ) );
    printk( "RAG = %d \n", adc18Ptr->RAG );
    printk( "BCR = 0x%x\n", adc18Ptr->BCR );
    adc18Ptr->RAG &= ~( GSAF_SAMPLE_START ); //  0x10000
    /// Initiate board calibration
    adc18Ptr->BCR |= GSAF_AUTO_CAL;
    /// Wait for internal calibration to complete.
    do
    {
        autocal++;
        udelay( 100 );
    } while ( ( adc18Ptr->BCR & GSAF_AUTO_CAL ) != 0 ); // 0x2000
    if ( ( adc18Ptr->BCR & GSAF_AUTO_CAL_PASS ) == 0 )
    {
        printk( "ADC AUTOCAL FAIL %d\n", autocal );
        autocal = 0;
    }
    else
    {
        printk( "ADC AUTOCAL PASS %d\n", autocal );
        autocal = GSAF_AUTO_CAL_PASS;
    }
    // End AutoCal

    adc18Ptr->RAG |= GSAF_SAMPLE_START; // 0x10000
    adc18Ptr->IDBC = ( GSAI_CLEAR_BUFFER | GSAI_THRESHOLD | GSAF_18BIT_DATA );
    // Determine number of channels on board
    pHardware->adcConfig[ devNum ] = adc18Ptr->ASSC;
    chans = ( pHardware->adcConfig[ devNum ] >> 16 ) & 0x3;
    // Set SSC register according to channel count ie all avail channels armed
    switch ( chans )
    {
    case 0:
        pHardware->adcChannels[ devNum ] = 32;
        adc18Ptr->SSC =
            ( GSAF_32_CHANNEL ); // Sets 32 channels and external clock
        break;
    case 1:
        pHardware->adcChannels[ devNum ] = 16;
        adc18Ptr->SSC =
            ( GSAF_16_CHANNEL ); // Sets 16 channels and external clock
        break;
    case 2:
    default:
        pHardware->adcChannels[ devNum ] = 8;
        adc18Ptr->SSC =
            ( GSAF_8_CHANNEL ); // Sets 8 channels and external clock
        break;
    }
#ifdef USE_ADC_CLOCK
    adc18Ptr->SSC |= GSAF_USE_RAG_CLK;
#endif
    printk( "SSC = 0x%x\n", adc18Ptr->SSC );
    printk( "IDBC = 0x%x\n", adc18Ptr->IDBC );
    /// Fill in CDS_HARDWARE structure with ADC information.
    pHardware->pci_adc[ devNum ] =
        (long)pci_alloc_consistent( adcdev, 0x2000, &adc_dma_handle[ devNum ] );
    pHardware->adcType[ devNum ] = GSC_18AI32SSC1M;
    pHardware->adcInstance[ devNum ] = pHardware->adc18Count;
    pHardware->adc18Count++;
    pHardware->adcConfig[ devNum ] |= autocal;
    pHardware->adcCount++;
    /// Return board enable status.
    return ( pedStatus );
}

// *****************************************************************************
/// \brief Function clears ADC buffer and starts acquisition via external clock.
///< Also sets up ADC for Demand DMA mode and set GO bit in DMA Mode Register.
///< NOTE: In normal operation, this code should only be called while the clocks
///< from the timing receiver are turned OFF ie during initialization process.
///	@param[in] adcCount Total number of ADC modules to start DMA.
// *****************************************************************************
void
gsc18ai32Enable( CDS_HARDWARE* pHardware )
{
    int ii;

    for ( ii = 0; ii < pHardware->adcCount; ii++ )
    {
        if ( pHardware->adcType[ ii ] == GSC_18AI32SSC1M )
        {
            /// Clear the FIFO and set 18bit data mode
            _adcPtr[ ii ]->IDBC =
                ( GSAI_CLEAR_BUFFER | GSAI_THRESHOLD | GSAF_18BIT_DATA );
            /// Set DMA mode and direction in PLX controller chip on module.
            /// Enable DMA
            plx9056_dma_enable( ii );
#ifdef USE_ADC_CLOCK
            // Start internal clock generator
            _adcPtr[ ii ]->RAG &= ~( GSAF_SAMPLE_START ); //  0x10000
#endif
            /// Enable sync via external clock input.
            _adcPtr[ ii ]->SSC |= GSAF_ENABLE_CLOCK;
        }
    }
}

// *****************************************************************************
/// \brief Routine reads number of samples in ADC FIFO.
///	@param[in] numAdc The ID number of the ADC module to read.
///	@return The number of samples presently in the ADC FIFO.
// *****************************************************************************
int
gsc18ai32CheckAdcBuffer( CDS_HARDWARE* pHardware, int numAdc )
{
    int dataCount = 0;

    if ( pHardware->adcType[ numAdc ] == GSC_18AI32SSC1M )
    {
        dataCount = _adcPtr[ numAdc ]->BUF_SIZE;
    }
    return ( dataCount );
}
