inline int iop_adc_init( adcInfo_t* );
inline int iop_adc_read( adcInfo_t*, int[] );
void iop_enable_all_adc_dac ( );

// Routine for initializing ADC modules on startup
inline int
iop_adc_init( adcInfo_t* adcinfo )
{
    volatile int* adcDummyData;
    int           status;
    int           ii, jj;
    int           dma_bytes = 32;

    /// \> If IOP,  Initialize the ADC modules
    for ( jj = 0; jj < cdsPciModules.adcCount; jj++ )
    {

        // Preload input memory with dummy variables to test that new ADC data
        // has arrived.
        // Write a dummy 0 to first ADC channel location
        // This location should never be zero when the ADC writes data as it
        // should always have an upper bit set indicating channel 0.
        // Write a number into the last channel which the ADC should never write
        // ie no upper bits should be set in channel 31.
        adcDummyData = (int*)cdsPciModules.pci_adc[ jj ];
        switch ( cdsPciModules.adcType[ jj ] )
        {
        case GSC_18AI32SSC1M:
            dma_bytes = cdsPciModules.adcChannels[ jj ] * 4 * UNDERSAMPLE;
            plx9056_dma_setup( jj , dma_bytes);
            *adcDummyData = 0x0;
            adcDummyData += (cdsPciModules.adcChannels[ jj ] * UNDERSAMPLE - 1);
            *adcDummyData = DUMMY_ADC_VAL;
            break;
        case GSC_18AI64SSC:
            dma_bytes = GSA7_DMA_BYTE_COUNT;
            plx9056_dma_setup( jj , dma_bytes);
            *adcDummyData = 0x0;
            adcDummyData += GSA7_64_OFFSET;
            *adcDummyData = DUMMY_ADC_VAL;
            break;
        default:
            dma_bytes = GSAI_DMA_BYTE_COUNT;
            plx9056_dma_setup( jj , dma_bytes);
            *adcDummyData = 0x0;
            adcDummyData += GSAI_64_OFFSET;
            *adcDummyData = DUMMY_ADC_VAL;
            break;
        }

        // Set ADC Present Flag
        pLocalEpics->epicsOutput.statAdc[ jj ] = ADC_MAPPED;
        // Set ADC AutoCal Pass/Fail Flag
        if ( ( cdsPciModules.adcConfig[ jj ] & GSAI_AUTO_CAL_PASS ) != 0 )
            pLocalEpics->epicsOutput.statAdc[ jj ] |= ADC_CAL_PASS;
        // Reset Diag Info
        adcinfo->adcRdTimeErr[ jj ] = 0;
        adcinfo->adcChanErr[ jj ] = 0;
        adcinfo->adcOF[ jj ] = 0;
        for ( ii = 0; ii < MAX_ADC_CHN_PER_MOD; ii++ )
            adcinfo->overflowAdc[ jj ][ ii ] = 0;
    }
    adcinfo->adcHoldTime = 0;
    adcinfo->adcHoldTimeMax = 0;
    adcinfo->adcHoldTimeEverMax = 0;
    adcinfo->adcHoldTimeEverMaxWhen = 0;
    adcinfo->adcHoldTimeMin = 0xffff;
    adcinfo->adcHoldTimeAvg = 0;
    adcinfo->adcWait = 0;
    adcinfo->chanHop = 0;

    return 0;
}

void 
iop_enable_all_adc_dac ( )
{
    /// - ---- Arm ADC modules
    gsc16ai64Enable( &cdsPciModules );
    gsc18ai32Enable( &cdsPciModules );
    gsc18ai64Enable( &cdsPciModules );
     /// - ----  Arm DAC outputs
    gsc18ao8Enable( &cdsPciModules );
    gsc20ao8Enable( &cdsPciModules );
    gsc16ao16Enable( &cdsPciModules );

    udelay( MAX_UDELAY );
    udelay( MAX_UDELAY );
}

// ADC Read *****************************************************************
// Routine for reading data from ADC modules.
inline int
iop_adc_read( adcInfo_t* adcinfo, int cpuClk[] )
{
    int           ii, kk;
    volatile int* packedData;
    int           limit;
    int           mask;
    unsigned int  offset;
    int           num_outs;
    int           ioMemCntr = 0;
    int           card;
    int           chan;
    int           loops = 1;
    int           iocycle = 0;
    unsigned int  nsec = 0;
    int           gps_status = 0;
    int           first_chan_marker = 0;
    int           adc_wait_limit = 0;

    // On startup, may take up to 1.25 seconds before clocks start
    // while waiting for 1PPS sync marker, so wait longer for
    // ADC data to arriver.
    if ( adcinfo->startup_sync )
        adc_wait_limit = MAX_ADC_WAIT;
    else
        adc_wait_limit = MAX_ADC_WAIT_CARD_0;

    // Read ADC data
    for ( card = 0; card < cdsPciModules.adcCount; card++ )
    {
        /// - ---- ADC DATA RDY is detected when last channel in memory no
        /// longer contains the dummy variable written during initialization and
        /// reset after each read.
        packedData = (int*)cdsPciModules.pci_adc[ card ];

        /// Where last channel is is dependent on number of channels a
        /// particular card type has.
        switch ( cdsPciModules.adcType[ card ] )
        {
        case GSC_18AI32SSC1M:
            packedData += (cdsPciModules.adcChannels[ card ] * UNDERSAMPLE - 1);
            first_chan_marker = GSAF_FIRST_SAMPLE_MARK;
            break;
        case GSC_18AI64SSC:
            packedData += GSA7_64_OFFSET;
            first_chan_marker = GSA7_FIRST_SAMPLE_MARK;
            break;
        default:
            packedData += GSAI_64_OFFSET;
            first_chan_marker = GSAI_FIRST_SAMPLE_MARK;
            break;
        }
        // Capture CPU clock cycle for timing diags
        cpuClk[ CPU_TIME_RDY_ADC ] = rdtsc_ordered( );
        // Wait for ADC data to arrive via DMA
        do
        {
            /// - ---- Need to delay if not ready as constant banging of the
            /// input register will slow down the ADC DMA.
            cpuClk[ CPU_TIME_ADC_WAIT ] = rdtsc_ordered( );
            adcinfo->adcWait =
                ( cpuClk[ CPU_TIME_ADC_WAIT ] - cpuClk[ CPU_TIME_RDY_ADC ] ) /
                CPURATE;
        } while ( ( *packedData == DUMMY_ADC_VAL ) &&
                  ( adcinfo->adcWait < adc_wait_limit ) );

        /// - ---- Added ADC timing diagnostics to verify timing consistent and
        /// all rdy together.
        if ( card == 0 )
        {
            adcinfo->adcRdTime[ card ] = ( cpuClk[ CPU_TIME_ADC_WAIT ] -
                                           cpuClk[ CPU_TIME_CYCLE_START ] ) /
                CPURATE;
#ifdef XMIT_DOLPHIN_TIME
            // Send time on Dolphin net if this is the time xmitter.
            pcieTimer->gps_time = timeSec;
            pcieTimer->cycle = cycleNum;
            clflush_cache_range( (void*)&pcieTimer->gps_time, 16 );
#endif
        }
        else
        {
            adcinfo->adcRdTime[ card ] = adcinfo->adcWait;
        }

        if ( adcinfo->adcRdTime[ card ] > adcinfo->adcRdTimeMax[ card ] )
            adcinfo->adcRdTimeMax[ card ] = adcinfo->adcRdTime[ card ];

        if ( ( card == 0 ) &&
             ( adcinfo->adcRdTimeMax[ card ] >
               ( MAX_ADC_WAIT_CARD_0 * UNDERSAMPLE ) ) )
            adcinfo->adcRdTimeErr[ card ]++;

        if ( ( card != 0 ) &&
             ( adcinfo->adcRdTimeMax[ card ] > MAX_ADC_WAIT_CARD_S ) )
            adcinfo->adcRdTimeErr[ card ]++;

        /// - --------- If data not ready in time, abort.
        /// Either the clock is missing or code is running too slow and ADC FIFO
        /// is overflowing.
        if ( adcinfo->adcWait >= adc_wait_limit )
        {
            pLocalEpics->epicsOutput.statAdc[ card ] &= ~( ADC_RD_TIME );
            fe_status_return = ADC_TO_ERROR;
            fe_status_return_subcode = card;
            return ( ADC_TO_ERROR );
        }

        if ( card == 0 ) // Capture timing info
        {
            // Capture start of cycle time for cpu meter diagnostics
            cpuClk[ CPU_TIME_CYCLE_START ] = rdtsc_ordered( );
            /// \> If first cycle of a new second, capture time from IRIG-B
            /// module or LIGO PCIe timing receiver.
            if ( cycleNum == 0 )
            {
                // if SymCom type, just do write to lock current time and read
                // later This save a couple three microseconds here
                switch ( cdsPciModules.gpsType )
                {
                case SYMCOM_RCVR:
                    lockGpsTime( );
                    break;
                case TSYNC_RCVR:
                    /// - ---- Reading second info will lock the time register,
                    /// allowing nanoseconds to be read later (on next cycle).
                    /// Two step process used to save CPU time here, as each
                    /// read can take 2usec or more.
                    timeSec = getGpsSecTsync( );
                    break;
                case LIGO_RCVR:
                    // Reading fractional time from LPTC will lock GPS second register
                    // for read on next cycle.
                    // Two step process used to save CPU time here, as each
                    // read can take 2usec or more.
                    timeSec = lptc_get_gps_sec( &cdsPciModules );
                    break;
                default:
                    break;
                }
            }
        }

        /// \> Read adc data
        // Set pointer to first channel
        packedData = (int*)cdsPciModules.pci_adc[ card ];
        /// - ---- First, and only first, channel should have upper bit marker
        /// set. If not, have a channel hopping error.
        if ( !( *packedData & first_chan_marker ) )
        {
            adcinfo->chanHop = 1;
            adcinfo->adcChanErr[ card ] = 1;
            fe_status_return = CHAN_HOP_ERROR;
            fe_status_return_subcode = card;
            return CHAN_HOP_ERROR;
        }

        // Various ADC models have different number of channels/data bits
        if ( cdsPciModules.adcType[ card ] == GSC_16AI64SSA )
        {
            limit = OVERFLOW_LIMIT_16BIT;
            offset = GSAI_DATA_CODE_OFFSET;
            mask = GSAI_DATA_MASK;
        }
        else
        {
            limit = OVERFLOW_LIMIT_18BIT;
            offset = GSA7_DATA_CODE_OFFSET;
            mask = GSA7_DATA_MASK;
        }
        num_outs = cdsPciModules.adcChannels[ card ];
        // If ADC clocking faster than code, then need to do multiple reads/code
        // cycle
        loops = UNDERSAMPLE;
        // GSC_16AI64SSA can't support rate > 128KS/sec
        if ( cdsPciModules.adcType[ card ] == GSC_16AI64SSA && UNDERSAMPLE > 4 )
            loops = 1;
        /// - ---- Determine next ipc memory location to load ADC data
        ioMemCntr = ( ( cycleNum / ADC_MEMCPY_RATE ) % IO_MEMORY_SLOTS );
        iocycle = ( cycleNum / ADC_MEMCPY_RATE );

        /// - ----  Read adc data from PCI mapped memory into local variables
        for ( kk = 0; kk < loops; kk++ )
        {
            for ( chan = 0; chan < num_outs; chan++ )
            {
                // adcData is the integer representation of the ADC data
                adcinfo->adcData[ card ][ chan ] = ( *packedData & mask );
                adcinfo->adcData[ card ][ chan ] -= offset;
#ifdef DEC_TEST
                // This only used on test system for checking decimation filters
                // by providing an ADC signal via a GDS EXC signal
                if ( chan == 0 )
                {
                    adcinfo->adcData[ card ][ chan ] =
                        dspPtr[ 0 ]->data[ 0 ].exciteInput;
                }
#endif
                // dWord is the double representation of the ADC data
                // This is the value used by the rest of the code calculations.
                dWord[ card ][ chan ][ kk ] = adcinfo->adcData[ card ][ chan ];
                // If running with fast ADC at 512K and gsc16ai64 at 64K, then:
                // fill in the missing data by doing data copy
                if ( cdsPciModules.adcType[ card ] == GSC_16AI64SSA &&
                     UNDERSAMPLE > 4 )
                {
                    for ( ii = 1; ii < UNDERSAMPLE; ii++ )
                        dWord[ card ][ chan ][ ii ] =
                            adcinfo->adcData[ card ][ chan ];
                }
                /// - ----  Load ADC value into ipc shared memory buffer
                ioMemData->iodata[ card ][ ioMemCntr ].data[ chan ] =
                    adcinfo->adcData[ card ][ chan ];
                /// - ---- Check for ADC overflows
                if ( ( adcinfo->adcData[ card ][ chan ] > limit ) ||
                     ( adcinfo->adcData[ card ][ chan ] < -limit ) )
                {
                    adcinfo->overflowAdc[ card ][ chan ]++;
                    pLocalEpics->epicsOutput.overflowAdcAcc[ card ][ chan ]++;
                    overflowAcc++;
                    adcinfo->adcOF[ card ] = 1;
                    odcStateWord |= ODC_ADC_OVF;
                }
                packedData++;
            }

            // For normal IOP ADC undersampling ie not a mixed fast 512K adc and
            // normal 64K adc
            if ( UNDERSAMPLE < 5 )
            {
                /// - ---- Write GPS time and cycle count as indicator to
                /// control app that adc data is ready
                ioMemData->gpsSecond = timeSec;
                ioMemData->iodata[ card ][ ioMemCntr ].timeSec = timeSec;
                ioMemData->iodata[ card ][ ioMemCntr ].cycle = iocycle;
                ioMemCntr++;
                iocycle++;
                iocycle %= IOP_IO_RATE;
            }
        }

        // For ADC undersampling when running with a fast ADC at 512K
        // to limit rate to user apps at 64K
        if ( UNDERSAMPLE > 4 )
        {
            /// - ---- Write GPS time and cycle count as indicator to control
            /// app that adc data is ready
            ioMemData->gpsSecond = timeSec;
            ioMemData->iodata[ card ][ ioMemCntr ].timeSec = timeSec;
            ioMemData->iodata[ card ][ ioMemCntr ].cycle = iocycle;
        }

        /// - ---- Clear out 1st channel from last ADC data read for test on
        /// next cycle
        /// - ---- First channel should never be zero ie should have channel
        /// marker bit
        packedData = (int*)cdsPciModules.pci_adc[ card ];
        *packedData = 0x0;

        // Enable the ADC Demand DMA for next read
        switch ( cdsPciModules.adcType[ card ] )
        {
        case GSC_18AI32SSC1M:
            packedData += (cdsPciModules.adcChannels[ card ] * UNDERSAMPLE - 1);
            *packedData = DUMMY_ADC_VAL;
            plx9056_dma_start( card );
            break;
        case GSC_18AI64SSC:
            packedData += GSA7_64_OFFSET;
            *packedData = DUMMY_ADC_VAL;
            plx9056_dma_start( card );
            break;
        default:
            packedData += GSAI_64_OFFSET;
            *packedData = DUMMY_ADC_VAL;
            plx9056_dma_start( card );
            break;
        }

#ifdef DIAG_TEST
        // For DIAGS ONLY !!!!!!!!
        // This will change ADC DMA BYTE count
        // -- Greater than normal will result in channel hopping.
        // -- Less than normal will result in ADC timeout.
        // In both cases, real-time kernel code should exit with errors to dmesg
        if ( pLocalEpics->epicsInput.bumpAdcRd != 0 )
        {
            gsc16ai64DmaBump( card, pLocalEpics->epicsInput.bumpAdcRd );
            pLocalEpics->epicsInput.bumpAdcRd = 0;
        }
#endif
        adc_wait_limit = MAX_ADC_WAIT_CARD_0;
    }
    return 0;
}
